import java.util.*;
import java.io.*;

// This file is included in XmlConverterP.java 
// abstract class TextConverter {
// 	XmlParentFormHandler handler = null;
//
// 	abstract void flush();
// }

class SimpleTextConverter extends TextConverter {

	SimpleTextConverter(XmlParentFormHandler handler) {
		this.handler = handler;
	}

	void flush() {
		String text = this.handler.textBuffer.toString(); 
		text = text.replaceAll("\n", "\\n");
		text = text.replaceAll("\r", "\\r");
		this.handler.writeStartAnAttrElem(text, "TEXT");
		this.handler.writeEndAnAttrElem();
		this.handler.textBuffer.setLength(0);
	}
}

class WordSplitTextConverter extends TextConverter {

	WordSplitTextConverter(XmlParentFormHandler handler) {
		this.handler = handler;
	}

	void flush() {
		StreamTokenizer in
			= new StreamTokenizer(
				  new StringReader(this.handler.textBuffer.toString()));
		in.ordinaryChar('#');

		try {
			while (in.nextToken() != StreamTokenizer.TT_EOF) {
				if (in.ttype == StreamTokenizer.TT_WORD) {
					this.handler.writeStartAnAttrElem(in.sval, "TEXT");
					this.handler.writeEndAnAttrElem();

					//this.out.println("### "+in.sval);
				}
				else if (in.ttype == StreamTokenizer.TT_NUMBER) {
					this.handler.writeStartAnAttrElem(String.valueOf((int)in.nval), "TEXT");
					this.handler.writeEndAnAttrElem();
					//this.out.println("### "+in.sval);
				}
				else if (in.ttype == StreamTokenizer.TT_EOL) {
					//this.out.println("### EOL");
				}
			}
			//this.out.println("### EOF\n");
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}

		this.handler.textBuffer.setLength(0);
	};
}

//EOF
