/*
 * @(#)Unot.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.util.*;

/**
 * Unot.java
 *
 *
 * Created: 13 May 2005
 *
 * @author <a href="mailto:arim@ist.hokudai.ac.jp">Hiroki Arimura</a>
 * @version $Revision: 1.5 $
 * @see <a href="http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf">
 * Discovering Frequent Substructures in Large Unordered Trees (PDF)</a>
 */
public class Match
{

    /** Java  DEBUG ͡ */
    public static final boolean DEBUG = false;
    public static final boolean DEBUG_UPDATE = false;
    public static final boolean DEBUG_OPTION = false;
    public static final boolean DEBUG_CLOSED = false;
    public static final boolean DEBUG_MATCH  = false;

    /** ｪλ֤ */
    public static final int EXIT_SUCCESS = 0;

    /** ｪλ֤ */
    public static final int EXIT_FAILURE = 1;

    /** Сʸ*/
    public static final String VERSION_STRING = "Match $Revision: 1.5 $";

    public static final String USAGE_STRING =
	"Usage: java [Java OPTIONS] unot.PatSet [OPTIONS...] PATFILE [PATFILES...] \n" +
	"ARGS \n" +
	" FILE [FILES]:\t input pattern iles (depth format).\n" +
	"OPTIONS \n" +
	" -pp | --printpatset \t print pattern set after match (default). \n" +
	" -pm | --printmaximal\t print maximal patterns only \n" +
	" -pc | --printmaximal\t print closed patterns only \n" +
	" -pd | --printdocset \t print pattern set. This put off -pp \n" +
	" -i  | --inputdocset \t match patterns against a given docset (parent format). \n" +
	" -ni | --inputdocset \t match patterns against patterns themselves (default) \n" +
	" -h  | --help     \t print this help. \n"+
	"";



    /** Υ饹Ѥɥȥåȡ󥹥ȥ饯ǻꤹ롥 */
    protected DocSet docSet;

	/** ѡ٤ */
	protected OccListTable occListTable = null; 

	/** ץ󡧥ѥ˺иޤ뤫ɤ */
	public static boolean useRootCount= true; 

	/** ץ󡧥ѥ˺иꥹȾޤ뤫ɤ */
	public static boolean useRootList = true; 

	/** ץ󡧥ѥʸиꥹȾޤ뤫ɤ */
	public static boolean useDocList = true; 

    /**
     * Creates a new <code>Match</code> instance.
     *
     * @param docSet a <code>DocSet</code> value
     */
    public Match(DocSet docSet)
    {
        if (docSet == null)
            throw new IllegalArgumentException("Error: DocSet is null.");

        this.docSet = docSet;

        // иꥹȤΥꥹȤ롥
		// η׻ϡ٤ƤΥѥ̤1Ǥ褤
		if (this.occListTable == null)
			this.occListTable = this.docSet.getOccLists();

    } // Match constructor


    /**
     * Executes `UNOT' main algorithm.
     *
     * @param pattern an <code>Pattern</code> value
     * @return a <code>OccList</code> value
     */
    public NodeList match(Pattern pattern)
    {
		if (DEBUG_MATCH) System.out.println("debug docSet.docList: "+
												docSet.getDocList().getNodeIdListString());

		if (DEBUG_MATCH) 
			System.out.println("\n\n@match pattern: "+
							   pattern.writeToString());

		if (pattern.getCodeSize() <= 0)
			return null;

		//ڥѥȹ
		OccList occList = matchOccList(pattern);

		NodeList rootList = occList.getRootList(); 
        if (DEBUG) UnotLog.log("@log NumOfRootOccs  : "+rootList.size() + "\n");

		if (useRootCount) {
			pattern.setRootCount(rootList.size());
		}

		if (useRootList) {
			pattern.setRootList(rootList);
		}

		if (useDocList) {
			pattern.setDocList(occList.getDocIdList());
		}


        return rootList;
    }

    OccList matchOccList(Pattern pattern)
    {
		MatchPattern workPattern = new MatchPattern();
		OccList occList = null; 

		if (DEBUG_MATCH) System.out.println("@matchOccList"+
											 " id="+pattern.getId()+"\n");
		for (int i = 0; i < pattern.getCodeSize(); i++) {
			DLpair pair = (DLpair) pattern.getNodeAt(i);
			workPattern.add(pair);
			if (DEBUG_MATCH) System.out.println("debug  index["+i+"] "+
												"id="+pattern.getId());
			if (i == 0) {
				int rootLabelId = pair.getLabel();
				occList = (OccList) this.occListTable.getOccList(rootLabelId);
			}
			else {
				occList = updateOcc(workPattern, occList);
			}
			if (DEBUG_MATCH) System.out.println("debug  --occList: "+
												occList.getRootList().getNodeIdListString());

		}
		return occList;
	}

	//expand
	//updateocc

	//Utility
	String getLabelStr(int id) {
		return this.docSet.getLabelDict().lookupLabel(id);
	}
						

    /**
     * Updates the old Occurrence List to the current Occurrence List
     * with the expanded <code>UnotPattern</code>.
     *
     * @param newPattern an <code>UnotPattern</code> value
     * @param oldOccList a <code>List</code> value
     * @return a <code>List</code> value
     */
    protected OccList updateOcc(MatchPattern newPattern, OccList oldOccList)
    {

        OccList newOccList = new OccList();

        int topDepth = newPattern.getTopDepth(); //The largest index of the right branch
        int topLabel = newPattern.getNodeLabel(topDepth);
        int parentIndex = newPattern.getNodeIndex(topDepth -1);

		if (DEBUG_UPDATE) System.out.println("debug: #update "+
											 "label "+topLabel+" "+getLabelStr(topLabel));

        for (int i = 0; i < oldOccList.size(); i++) {
            Occurrence occ = (Occurrence) oldOccList.get(i);
            Occurrence newOcc = null;

            UnotNode parent = (UnotNode) occ.get(parentIndex);
            NodeList children = parent.getChildren();

			if (DEBUG_UPDATE) System.out.println("debug: parent"+i+" "+parent+
												 " "+getLabelStr(parent.getLabelId()));

            for (int j = 0; j < children.size(); j++) {
                UnotNode child = (UnotNode) children.get(j);
				if (DEBUG_UPDATE) System.out.println("debug:   child"+j+" "+child+
												 " "+getLabelStr(child.getLabelId()));
                if ((child.getLabelId() == topLabel) && ! occ.contains(child)) {
                    newOcc = new Occurrence(occ, child);
					newOccList.add(newOcc);
					if (DEBUG_UPDATE) System.out.println("debug:     succeeded!");
				}
				else {
					if (DEBUG_UPDATE) System.out.println("debug:     failed!");
				}
            }
        }

        return newOccList;

    } // updateOcc


	/*
	 *
	 */

	void markMatchList(Pattern pattern, PatSet patSet) {
		patSet.ensureNonSizeOrdered();
			
		IndexList docList = pattern.getDocList();
		for (int i = 0; i < docList.size(); i++) {
			//ա󤢤ꡪ
			int patId = docList.getValue(i);
			Pattern matched = patSet.getPattern(patId);
			matched.markMatch(pattern.getId());
		}
	}

	////////////////////////////////////////////

	public static boolean agree(IndexList A, IndexList B) {
		if (A.size() != B.size())
			return false;
		for (int i = 0; i < A.size(); i++) {
			if (A.getValue(i) != B.getValue(i)) {
				if (DEBUG_CLOSED) System.out.println("@debug: disagree at "+i+":"+
								   A.getValue(i)+"!="+B.getValue(i));
				return false;
			}
		}
		if (DEBUG_CLOSED) System.out.println("@debug: agree!");
		return true;
	}

	/** Check if a given pattern is maximal (globally maximal) */
	static boolean isMaximal(Pattern pattern) {
		return pattern.getDocList().size() <= 1; 
	}

	/** Check if a given pattern is closed (position maximal) */
	static boolean isPositionMaximal(Pattern pattern, PatSet patSet) {
		patSet.ensureNonSizeOrdered();

		if (DEBUG_CLOSED) System.out.println("===== enter isPositionMaximal: "+pattern.getId());
			
		IndexList docList = pattern.getDocList();
		for (int i = 0; i < docList.size(); i++) {
			int patId = docList.getValue(i);
			//ա󤢤ꡪ

			if (DEBUG_CLOSED) System.out.println("debug check matched: "+patId);

			if (patId == pattern.getId()) {
				if (DEBUG_CLOSED) System.out.println("debug   skip: to self");
				continue; //ʬ. skip
			}

			Pattern matchedPattern = patSet.getPattern(patId);

			IndexList rootList0 = pattern.orgRootList; 
			IndexList rootList1 = matchedPattern.orgRootList;

			if (rootList0.size() != rootList1.size()) {
				if (DEBUG_CLOSED) System.out.println("debug   skip: diff count");
				continue; //skip this matchedPattern
			}

				if (DEBUG_CLOSED) System.out.println("debug   check rootlist");

			if (agree(rootList0, rootList1)) {
				//pattern and matchedPatterns have the same rootLists
				if (DEBUG_CLOSED) System.out.println("debug     nonclsoed: same rootlist");
				return false; 
			}
		}
		if (DEBUG_CLOSED) System.out.println("debug   closed: ");
		return true;
	}

	static void writeFileClosedPatterns(PatSet patSet) {
			PatternWriter writer = Pattern.getPatternWriter();
			int numMax  = 0;
			int numSkip = 0;
			System.out.println("@writer "+Pattern.getPatternWriter().getName()+"\n");
			for (int i = 0; i < patSet.size(); i++) {
				Pattern pattern = patSet.get(i);
				if (isPositionMaximal(pattern, patSet)) {
					System.out.println(writer.write(pattern)+"");
					numMax++;
				}
				else {
					numSkip++;
				}
			}
			UnotLog.log("@log MaxPat:" +
						"\tnumClosed "+numMax+
						"\tnumSkipped "+numSkip+
						"\tnumTotal   "+patSet.size()+
						"\n");
	}

	static void writeFileMaxPatterns(PatSet patSet) {
			PatternWriter writer = Pattern.getPatternWriter();
			int numMax  = 0;
			int numSkip = 0;
			System.out.println("@writer "+Pattern.getPatternWriter().getName()+"\n");
			for (int i = 0; i < patSet.size(); i++) {
				Pattern pattern = patSet.get(i);
				if (isMaximal(pattern)) {
					System.out.println(writer.write(pattern)+"");
					numMax++;
				}
				else {
					numSkip++;
				}
			}
			UnotLog.log("@log MaxPat:" +
						" numMaximal "+numMax+
						"\tnumSkipped "+numSkip+
						"\tnumTotal   "+patSet.size()+
						"\n");
	}

	////////////////////////////////////////////

	/*
	 * Static Fields Option
	 */
	static boolean usePrintPatSet = true;  //default
	static boolean usePrintDocSet = false; 
	static boolean usePrintOnly   = false; 
	static boolean usePrintMaxPat = false; 
	static boolean usePrintClosedPat = false; 
	static String dataFileName = null; 

    static void read_options(String[] args) {

		while (argc < args.length){

			if (DEBUG_OPTION) System.out.println("debug: option "+args[argc]);

			if (args[argc].charAt(0) == '-') { //An option starts

				if (args[argc].equals("-h") || args[argc].equals("-help")){
					argc++;
					System.out.println(USAGE_STRING);
					System.exit(0);
				}
				else if (args[argc].equals("-pp") || args[argc].equals("-printpatset")){
					argc++;
					usePrintPatSet = true;
				}
				else if (args[argc].equals("-po") || args[argc].equals("-printonly")){
					argc++;
					usePrintOnly = true;
				}
				else if (args[argc].equals("-pd") || args[argc].equals("-printdocset")){
					argc++;
					usePrintDocSet = true;
					//usePrintPatSet = false;
				}
				else if (args[argc].equals("-pm") || args[argc].equals("-printmaximal")){
					argc++;
					usePrintMaxPat = true;
				}
				else if (args[argc].equals("-pc") || args[argc].equals("-printclosed")){
					argc++;
					usePrintClosedPat = true;
				}
				else if (args[argc].equals("-ni") || args[argc].equals("-noinputdatafile")){
					argc++;
					dataFileName = null;
				}
				else if (args[argc].equals("-i") ||
						 args[argc].equals("-inputdatafile")){
					argc++;
					dataFileName = args[argc++];
				}
				else if (args[argc].equals("-")) { //This indicates the stdin
					//Note: donot increment argc!
					break; //while loop
				}
				// Other case
				else {
					System.out.println("error: unknown option!: "+
									   args[argc]);
					System.exit(1);
				}
			}
			else {
				break;
			}

		}
	}

	static int argc = 0;

    /**
     * ƥ
     *
     * @param args[] a <code>String</code> value
     */
	public static void main(String args[])
	{
		///////////////
		Match.argc = 0;
		
        if (args.length < 1) {
            System.out.println(USAGE_STRING);
            System.exit(0);
        }

		Match.read_options(args);

		List<String> filenames = new ArrayList<String>();
        while (argc < args.length) {
			filenames.add(args[argc++]);
        }

		///
		
		DocSet docSet = new DocSet();
		docSet.setIndent(true);

		// Create an empty pattern set 
		PatSet patSet = new PatSet(docSet);
		patSet.setSizeOrdered(false); //Simple list
		Pattern.setPatternWriter(new DepthPatternWriter(docSet));
		Pattern.getPatternWriter().setIndent(true);
		if (usePrintMaxPat || usePrintClosedPat) {
			patSet.setRequireRootOcc(true);
		}

		//Load patterns from file
		for (int i = 0; i < filenames.size(); i++) {
			String filename = (String)filenames.get(i); 
			if (filename.equals("-")) {
				patSet.readFile(System.in); //STDIN
				break;
			}
				patSet.readFile(filename); //FILE
		}

		//DetSet preparation
		if (dataFileName == null) {
			//Match patterns agains patterns

			//Load patterns into DocSet
			patSet.loadPatterns(docSet);

		}
		else {
			//Match patterns agains dataset

			// Reading a data file
			if (dataFileName != null) {
				List<String> names = new ArrayList<String>();
				names.add(dataFileName);
				docSet.readFiles(names);
				UnotLog.log("@log DocSet: " + docSet + "\n");
			}
		}

		if (!usePrintOnly) {
			//Match patterns 
			Match matcher = new Match(docSet);
			for (int i = 0; i < patSet.size(); i++) {
				Pattern pattern = patSet.get(i);

				//Match pattern 
				NodeList rootList = matcher.match(pattern);

				//update matchList
				//matcher.markMatchList(pattern, patSet); 
			}
		}

		if (usePrintMaxPat) {
			System.out.println("@log print all maximal patterns");
			writeFileMaxPatterns(patSet);
		}
		else if (usePrintClosedPat) {
			System.out.println("@log print all closed patterns");
			writeFileClosedPatterns(patSet);
		}
		else if (usePrintPatSet) { 		//Print patterns
			System.out.println("@log print all patterns");
			patSet.writeFile(System.out);
		}

 		//Print documents
		if (usePrintDocSet) {
			System.out.println("\n"+"@log print docSet");
			docSet.writeFile(System.out);
		}

		return;
	}


} // Match
