/*
 * @(#)DocSet.java
 *
 * Copyright 2005, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.util.ArrayList;

/**
 * A table of OccLists indexed by their labels
 * @author Hiroki Arimura 
 * @version $Revision$
 */
public class OccListTable extends ArrayList<OccList>
{
	static boolean DEBUG_MATCH = false;

	OccListTable(int maxLabelSize) {
		super(maxLabelSize);

		// labelSize Ĥ OccList  occLists 
		for (int i = 0; i < maxLabelSize; i++) {
			this.add(new OccList());
		}
	}

	//void put(int labelId, UnotNode node) {
	void putOcc(int labelId, UnotNode node) {
		if (DEBUG_MATCH) System.out.println("@OccListTable: insert"+
											" labelId="+labelId+
											" node"+node);
		OccList occList = getOccList(labelId);
		Occurrence occ = new Occurrence();
		occ.add(node); 
		occList.add(occ);
	}

	OccList getOccList(int labelId) {
		if (labelId >= this.size()) {
			System.out.println("Error: getOccList: index out of bound: "+labelId);
			System.exit(1);
		}
		return (OccList) this.get(labelId);
	}

} // OccListTable


//EOF 

