/*
 * @(#)Pattern.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.io.*;
import java.util.*;

/**
 * Pattern.java
 *
 * Υ饹 UNOT ѥΥǡ¤ɽ롥 (trcs216.pdf pp.8 Fig.4)
 *
 * Created: Thu Aug  7 20:34:14 2003
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.5 $
 */
public class Pattern implements Cloneable
{
	/*
	 * Static Fields
	 */
    /** Pattern ֹ̤ ID */
	static private int _freePatId = 0;

	static int getPatId() {
		return _freePatId++;
	};

    /** ڤѤdocset֥ */
    static DocSet docSet = null;

    /** ڤϼ³ */
	static PatternReader reader = null;

    /** ڤϼ³ */
	static PatternWriter writer = null;

	/** ѥϥ롼ꤹ.	 */
	public static void setPatternReader(PatternReader reader) {
		Pattern.reader = reader; 
	};

	/** ѥϥ롼ꤹ.	 */
	public static void setPatternWriter(PatternWriter writer) {
		Pattern.writer = writer;
	};

	/** ѥϥ롼֤.	 */
	public static PatternWriter getPatternWriter() {
		return Pattern.writer; 
	};

    /** Infinity Node ؤβۥǥå
     *  left ޤ right ˤͤС˻ҶʤȤɽ
	 */
    public static final int INFTY_INDEX = -1;

	/*
	 * Member Fields
	 */

    /** (depth, label) Ф */
    protected DLpair[] code;

    /** code ˤ뱦ڤؤźǥåͤ */
    protected int[] right;

    /** ѥֹ̤.
	 * ѥȤ˰Ĥä롥
	 */
    protected int id = -1; //undef

    /** иȤʤΡɤθĿץ */
    protected int rootCount = -1; //undef

    /** иȤʤΡɤθĿץ */
    protected int docCount = -1; //undef

    /** иȤʤΡɤΥꥹȡץ */
    protected NodeList rootList = null;

    /** иʸIDΥꥹȡץ */
    protected IndexList docList = null; 

    /** Υѥ˾ȹ礹ѥIDΥꥹȡץ */
    protected IndexList matchList = null;

    /** иȤʤΡɤΥꥹȡMatch饹Τᡥ¸Ū*/
    public IndexList orgRootList = null;

    /**
     * Creates a new <code>Pattern</code> instance.
     *
     */
    public Pattern()
    {
        // do nothing.
		this.id = getPatId();
    } // Pattern constructor


    /**
     * Returns a Pattern object that has only root node.
     *
     * @param rootLabel an <code>int</code> value
     */
    public Pattern(int rootLabel)
    {
        this.code = new DLpair[] { new DLpair(0, rootLabel) };
        this.right = new int[] { INFTY_INDEX };
		this.id = getPatId();
    }


    /**
     * Creates a new <code>Pattern</code> instance.
     *
     * @param code a <code>DLpair[]</code> value
     * @param right an <code>int[]</code> value
     */
    public Pattern(DLpair[] code, int[] right)
    {
        this.code = code;
        this.right = right;
		this.id = getPatId();
    } // Pattern constructor

    /**
     * Get the code size (= the number of nodes).
     *
     * @return an <code>size</code> of the pattern
     */
    public int getCodeSize()
    {
        return this.code.length;
    }


    /**
     * Get the depth-label pair for the i-th node. 
     *
     * @return a DLpair <code>(depth, label)</code>
     */
    public DLpair getNodeAt(int i)
    {
        return this.code[i];
    }

    public int[] getRight() { return this.right;  }

    public int   getRootCount() { return this.rootCount; }
    public int   getDocCount() { return this.docCount; }
    public NodeList getRootList() { return this.rootList; }
    public IndexList getDocList() { return this.docList; }

    public void  setRootCount(int count) { this.rootCount = count; }
    public void  setDocCount(int count) { this.docCount = count; }
    public void  setRootList(NodeList list) { this.rootList = list; }
    public void  setDocList(IndexList list) { this.docList = list; }

    public IndexList getMatchList() { return this.matchList; }
    public void  setMatchList(IndexList list) { this.matchList = list; }
	/** Υѥ˥ޥåѥIDɲä. ץ */
	void markMatch(int matchPatId) {
		if (this.matchList == null)
			this.matchList = new IndexList();
		this.matchList.add(matchPatId);
	}


    public int   getId() { return this.id; }

    public void  setId(int id) { this.id = id; }

    /**
     * Get the top depth of the rightmost branch.
     *
     * @return an <code>int</code> value
     */
    public int getTopDepth()
    {
        return this.right.length - 1;
    }


    /**
     * Gets the top index of the `code' array of this object.
     *
     * @return an <code>int</code> value
     */
    public int getTopIndex()
    {
        return this.code.length - 1;
    }


    /**
     * Gets the index of the node on the `code' array of this object.
     *
     * @param depth an <code>int</code> value
     * @return an <code>int</code> value
     */
    public int getNodeIndex(int depth)
    {
        return (depth == 0) ? 0 : this.right[depth-1];
    }


    /**
     * Gets the label of the rightmost branch node at the depth.
     *
     * @param depth an <code>int</code> value
     * @return an <code>int</code> value
     */
    public int getNodeLabel(int depth)
    {
        int index = getNodeIndex(depth);

        if (index == INFTY_INDEX) {
            return -1; //error!
			//System.out.println("Error: Pattern.getNodeLabel: "+
			//"INFTY_INDEX!");
			//System.exit(1);
        } else {
            return this.code[index].getLabel();
        }
    }


    /**
     * Gets the index of the depth in the rightmost branch.
     *
     * @param depth an <code>int</code> value
     * @return an <code>int</code> value
     */
    public int getRightIndex(int depth)
    {
        return this.right[depth];
    }

    /**
     * Distinguish whether the depth node is Infinity Node or not.
     *
     * @param depth an <code>int</code> value
     * @return a <code>boolean</code> value
     */
    public boolean isInftyNode(int depth)
    {
        return (getNodeIndex(depth) == INFTY_INDEX) ? true : false;
    }

    /**
     * Returns a Pattern which is expanded with the Rightmost Expantion.
     *
     * @param oldPattern an <code>Pattern</code> value
     * @param newNode a <code>DLpair</code> value
     * @param succCopyDepth an <code>int</code> value
     * @return an <code>Pattern</code> value
     */
    public Pattern rightmostExpand(Pattern oldPattern, DLpair newNode, int succCopyDepth)
    {
        Pattern newPattern = new Pattern();

        int codeLength = oldPattern.code.length + 1;
        int rightLength = newNode.getDepth() + 1;

        // allocate
        newPattern.code = new DLpair[codeLength];
        newPattern.right = new int[rightLength];

        // copy
        System.arraycopy(oldPattern.code, 0, newPattern.code, 0, codeLength-1);
        System.arraycopy(oldPattern.right, 0, newPattern.right, 0, rightLength-1);

        int topIndex = newPattern.getTopIndex();
        int topDepth = newPattern.getTopDepth();

        // add new node
        newPattern.code[topIndex] = newNode;
        newPattern.right[topDepth-1] = topIndex;

        // add infinity node
        newPattern.right[topDepth] = INFTY_INDEX;

        return newPattern;

    } // rightmostExpand


    /**
     * Returns a cloned object.
     *
     * @return an <code>Object</code> value
     * @exception CloneNotSupportedException if an error occurs
     */
    public Object clone() throws CloneNotSupportedException
    {
        Pattern newPattern = (Pattern) super.clone();
        newPattern.code  = (DLpair[]) this.code.clone();
        newPattern.right = (int[]) this.right.clone();
        return (Object) newPattern;
    }


   /**
     * Retrieves a String representation of the object.
     *
     * @return a <code>String</code> representation of this object.
     * @see Object#toString()
     */
    public String toString()
    {
        StringBuffer strBuf = new StringBuffer();

        strBuf.append("[id: "+this.id);
        strBuf.append(" code=[");
		if (this.code != null)
			for (int i = 0; i < this.code.length; i++) {
				strBuf.append(this.code[i].toString());
			}
		else strBuf.append("null");
        strBuf.append("]");
        strBuf.append(" right=[");
		if (this.code != null)
			for (int i = 0; i < this.right.length; i++) {
				if (0 < i) strBuf.append(", ");
				strBuf.append(this.right[i]);
			}
		else strBuf.append("null");
        strBuf.append("]");
        strBuf.append("]");

        return strBuf.toString();
    }

    /**
     * Retrieve a tree like String representation of the object.
     *
     * @return a <code>String</code> value
     */
    public String writeToString() {
		return Pattern.writer.write(this);
	}

} // Pattern

////////////////////////////////////////////////


/**
 * A template for a reader class.
 * Each object of this class has the only method read(InputStream).
 */
abstract class PatternReader {

	abstract Pattern read();
}



//EOF
